use std::{
    io::{stdout, Write},
    sync::{mpsc::{self, Sender}, Arc, Mutex}, thread::{self, sleep}, time::Duration,
};

use ore::{self, state::{Bus, Hash}, BUS_ADDRESSES, BUS_COUNT, EPOCH_DURATION};
use rand::{rngs::OsRng, Rng};
use solana_program::{keccak::HASH_BYTES, program_memory::sol_memcmp, pubkey::Pubkey};
use solana_sdk::{
    compute_budget::ComputeBudgetInstruction,
    keccak::{hashv, Hash as KeccakHash},
    signature::Signer,
};
use tokio::sync::watch;

use crate::{
    cu_limits::{CU_LIMIT_MINE, CU_LIMIT_RESET},
    utils::{get_clock_account, get_proof, get_treasury},
    Miner,
};

// Odds of being selected to submit a reset tx
const RESET_ODDS: u64 = 20;

struct SharedState {
    proof_hash: Hash,
    treasury_difficulty: Hash,
}

impl Miner {
    pub async fn mine(self: Arc<Self>, threads: u64) {
        // Register, if needed.
        let signer = self.signer();
        self.register().await;
        // let mut stdout = stdout();
        
        // 创建一个通道，用于接收next_hash和nonce
        let (tx, rx) = mpsc::channel();
        let self_clone = self.clone();
        tokio::spawn(async move {
            self_clone.submit_solutions(rx).await;
        });

        // 创建watch channel
        let (tx_watch, rx_watch) = watch::channel(SharedState {
            proof_hash: Hash([0u8; HASH_BYTES]),
            treasury_difficulty: Hash([0u8; HASH_BYTES]),
        });

        // 启动状态监控
        let self_clone_monitor = self.clone();
        let rx_watch_for_state_monitoring = rx_watch.clone(); // 在传递给 tokio::spawn 之前克隆
        tokio::spawn(async move {
            self_clone_monitor.monitor_state_changes(tx_watch, rx_watch_for_state_monitoring).await;
        });


        // Start mining loop
        loop {
            // Fetch account state
            let balance = self.get_ore_display_balance().await;
            let treasury = get_treasury(&self.rpc_client).await;
            let proof = get_proof(&self.rpc_client, signer.pubkey()).await;
            let rewards =
                (proof.claimable_rewards as f64) / (10f64.powf(ore::TOKEN_DECIMALS as f64));
            let reward_rate =
                (treasury.reward_rate as f64) / (10f64.powf(ore::TOKEN_DECIMALS as f64));
            // stdout.write_all(b"\x1b[2J\x1b[3J\x1b[H").ok();
            println!("Balance: {} ORE", balance);
            println!("Claimable: {} ORE", rewards);
            println!("Reward rate: {} ORE", reward_rate);

            // Escape sequence that clears the screen and the scrollback buffer
            println!("\nMining for a valid hash...");
            let (next_hash, nonce) =
                self.find_next_hash_par(threads, tx.clone(), rx_watch.clone());

            // Submit mine tx.
            // Use busses randomly so on each epoch, transactions don't pile on the same busses
            println!("\n\nSubmitting hash for validation...");
            'submit: loop {
                // Double check we're submitting for the right challenge
                let proof_ = get_proof(&self.rpc_client, signer.pubkey()).await;
                if self.validate_hash(
                    next_hash,
                    proof_.hash.into(),
                    signer.pubkey(),
                    nonce,
                    treasury.difficulty.into(),
                ) {
                    println!("Hash validated! Submitting to the queue.");
                    // 如果验证成功，则发送next_hash和nonce到通道
                    if let Err(e) = tx.send((next_hash, nonce)) {
                        println!("Failed to send hash and nonce to the submit queue: {:?}", e);
                    }
                    break 'submit; // 成功提交后跳出循环
                } else {
                    println!("Hash already validated! An earlier transaction must have landed.");
                    break 'submit;
                }
            }
        }
    }

    async fn monitor_state_changes(self: Arc<Self>, tx_watch: tokio::sync::watch::Sender<SharedState>, mut rx_watch: tokio::sync::watch::Receiver<SharedState>) {
        println!("monitor_state_changes thread started");
        let mut is_first_update = true; // 标记是否为第一次更新
    
        loop {
            // 从服务器获取新的proof和treasury值
            let new_proof_hash = Hash::from(get_proof(&self.rpc_client, self.signer().pubkey()).await.hash);
            let new_treasury_difficulty = Hash::from(get_treasury(&self.rpc_client).await.difficulty);
    
            // 如果是第一次更新，输出proof
            if is_first_update {
                println!("Initial Proof Hash: {} Initial Treasury Difficulty: {}", new_proof_hash.to_string(), new_treasury_difficulty.to_string());
                is_first_update = false; // 更新标记，表示已经不是第一次更新了
            }
    
            // 如果有变化，则发送到共享状态
            if rx_watch.borrow().proof_hash != new_proof_hash || rx_watch.borrow().treasury_difficulty != new_treasury_difficulty {
                println!("New Proof Hash: {} New Treasury Difficulty: {}", new_proof_hash.to_string(), new_treasury_difficulty.to_string());
                tx_watch.send(SharedState {
                    proof_hash: new_proof_hash,
                    treasury_difficulty: new_treasury_difficulty,
                }).expect("Failed to send shared state update");
            }
    
            // 等待一段时间再次检查
            tokio::time::sleep(tokio::time::Duration::from_secs(5)).await;
        }
    }
    

    async fn submit_solutions(self: Arc<Self>, rx: mpsc::Receiver<(KeccakHash, u64)>) {
        println!("submit_solutions thread started");
        while let Ok((next_hash, nonce)) = rx.recv() {
            println!("Received hash: {} Nonce: {}", next_hash.to_string(), nonce);
            // Reset epoch, if needed
            // let mut rng = rand::thread_rng();
            let signer = self.signer();
            let treasury = get_treasury(&self.rpc_client).await;
            // let clock = get_clock_account(&self.rpc_client).await;
            // let threshold = treasury.last_reset_at.saturating_add(EPOCH_DURATION);
            // if clock.unix_timestamp.ge(&threshold) {
            //     // There are a lot of miners right now, so randomly select into submitting tx
            //     if rng.gen_range(0..RESET_ODDS).eq(&0) {
            //         println!("Sending epoch reset transaction...");
            //         let cu_limit_ix =
            //             ComputeBudgetInstruction::set_compute_unit_limit(CU_LIMIT_RESET);
            //         let cu_price_ix =
            //             ComputeBudgetInstruction::set_compute_unit_price(self.priority_fee);
            //         let reset_ix = ore::instruction::reset(signer.pubkey());
            //         self.send_and_confirm(&[cu_limit_ix, cu_price_ix, reset_ix], false, true)
            //             .await
            //             .ok();
            //     }
            // }
            // Submit request.
            let bus = self.find_bus_id(treasury.reward_rate).await;
            let bus_rewards = (bus.rewards as f64) / (10f64.powf(ore::TOKEN_DECIMALS as f64));
            println!("Sending on bus {} ({} ORE)", bus.id, bus_rewards);
            let cu_limit_ix = ComputeBudgetInstruction::set_compute_unit_limit(CU_LIMIT_MINE);
            let cu_price_ix =
                ComputeBudgetInstruction::set_compute_unit_price(self.priority_fee);
            let ix_mine = ore::instruction::mine(
                signer.pubkey(),
                BUS_ADDRESSES[bus.id as usize],
                next_hash.into(),
                nonce,
            );
            match self
                .send_and_confirm(&[cu_limit_ix, cu_price_ix, ix_mine], false, false)
                .await
            {
                Ok(sig) => {
                    println!("Success: {}", sig);
                }
                Err(err) => {
                    println!("Error: {:?}", err);
                }
            }
        }
        println!("submit_solutions thread ended");
    }

    async fn find_bus_id(&self, reward_rate: u64) -> Bus {
        let mut rng = OsRng;
        loop {
            let bus_id = rng.gen_range(0..BUS_COUNT);
            if let Ok(bus) = self.get_bus(bus_id).await {
                if bus.rewards.gt(&reward_rate.saturating_mul(20)) {
                    return bus;
                }
            }
        }
    }

    fn _find_next_hash(&self, hash: KeccakHash, difficulty: KeccakHash) -> (KeccakHash, u64) {
        let signer = self.signer();
        let mut next_hash: KeccakHash;
        let mut nonce = 0u64;
        loop {
            next_hash = hashv(&[
                hash.to_bytes().as_slice(),
                signer.pubkey().to_bytes().as_slice(),
                nonce.to_le_bytes().as_slice(),
            ]);
            if next_hash.le(&difficulty) {
                break;
            } else {
                println!("Invalid hash: {} Nonce: {:?}", next_hash.to_string(), nonce);
            }
            nonce += 1;
        }
        (next_hash, nonce)
    }


    fn find_next_hash_par(
        &self,
        // hash: KeccakHash,
        // difficulty: KeccakHash,
        threads: u64,
        tx: Sender<(KeccakHash, u64)>,
        rx_watch: watch::Receiver<SharedState>,
    ) -> (KeccakHash, u64) {
        let solution = Arc::new(Mutex::<(KeccakHash, u64)>::new((
            KeccakHash::new_from_array([0; 32]),
            0,
        )));
        let signer = self.signer();
        let pubkey = signer.pubkey();
        let thread_handles: Vec<_> = (0..threads)
            .map(|i| {
                std::thread::spawn({
                    let tx = tx.clone();
                    let solution = solution.clone();
                    // let mut stdout = stdout();
                    let rx_watch_clone = rx_watch.clone();
                    move || {
                        let n = u64::MAX.saturating_div(threads).saturating_mul(i);
                        let mut next_hash: KeccakHash;
                        let mut nonce: u64 = n;
                        loop {
                            // 这里，我们从rx_watch中获取最新的proof和difficulty
                            let current_state = rx_watch_clone.borrow();
                            let current_proof_hash = KeccakHash::from(current_state.proof_hash);
                            let current_difficulty = KeccakHash::from(current_state.treasury_difficulty);

                            next_hash = hashv(&[
                                current_proof_hash.to_bytes().as_slice(),
                                pubkey.to_bytes().as_slice(),
                                nonce.to_le_bytes().as_slice(),
                            ]);
                            // if nonce % 10_000 == 0 {
                            //     if n == 0 {
                            //         stdout
                            //             .write_all(
                            //                 format!("\r{}", next_hash.to_string()).as_bytes(),
                            //             )
                            //             .ok();
                            //     }
                            // }
                            if next_hash.le(&current_difficulty) {
                                println!("Task {} found solution and Sending hash: {} Nonce: {}", i, next_hash.to_string(), nonce);
                                let _ = tx.send((next_hash, nonce));
                                // let mut w_solution = solution.lock().expect("failed to lock mutex");
                                // *w_solution = (next_hash, nonce);
                                // return;
                            }
                            nonce += 1;
                        }
                    }
                })
            })
            .collect();

        for thread_handle in thread_handles {
            thread_handle.join().unwrap();
        }

        let r_solution = solution.lock().expect("Failed to get lock");
        *r_solution
    }

    pub fn validate_hash(
        &self,
        hash: KeccakHash,
        current_hash: KeccakHash,
        signer: Pubkey,
        nonce: u64,
        difficulty: KeccakHash,
    ) -> bool {
        // Validate hash correctness
        let hash_ = hashv(&[
            current_hash.as_ref(),
            signer.as_ref(),
            nonce.to_le_bytes().as_slice(),
        ]);
        if sol_memcmp(hash.as_ref(), hash_.as_ref(), HASH_BYTES) != 0 {
            return false;
        }

        // Validate hash difficulty
        if hash.gt(&difficulty) {
            return false;
        }

        true
    }

    pub async fn get_ore_display_balance(&self) -> String {
        let client = self.rpc_client.clone();
        let signer = self.signer();
        let token_account_address = spl_associated_token_account::get_associated_token_address(
            &signer.pubkey(),
            &ore::MINT_ADDRESS,
        );
        match client.get_token_account(&token_account_address).await {
            Ok(token_account) => {
                if let Some(token_account) = token_account {
                    token_account.token_amount.ui_amount_string
                } else {
                    "0.00".to_string()
                }
            }
            Err(_) => "0.00".to_string(),
        }
    }
}
